/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Interface to the cache manager.
 */

#include "arla_local.h"
RCSID("$Id: inter.c,v 1.35 1998/02/22 11:25:32 assar Exp $") ;

#include <xfs_message.h>

#if 0
/*
 * Return the rights for user cred and entry e.
 * The locking of e is up to the caller.
 */

u_long
getrights (FCacheEntry *e, CredCacheEntry *ce)
{
     AccessEntry *ae;

     if (!findaccess (ce->cred, e->acccache, &ae)) {
	 int ret;

	 ret = readattr(e, ce); /* XXX */
	 if (!findaccess (ce->cred, e->acccache, &ae))
	     ARLADEB (ADEBMISC, ("rightsokp: cannot get rights\n"));
     }
     return ae->access;
}

/*
 * Check to see if the operation(s) mask are allowed to user cred on
 * file e
 */

static Bool
checkright (FCacheEntry *e, u_long mask, CredCacheEntry *ce)
{
     return (e->anonaccess & mask) == mask
	  || (getrights (e, ce) & mask) == mask;
}
#endif

static Bool
checkright (FCacheEntry *e, u_long mask, CredCacheEntry *ce)
{
    return TRUE;
}

/*
 * These functions often take a FID as an argument to be general, but
 * they are intended to be called from a vnode-type of layer.
 */

/*
 * The interface to the open-routine.
 * This means that the file should be in the local cache and that the
 * cache manager should recall that some process has this file opened
 * for reading and/or writing.
 * We do no checking here. Should we?
 */

Result
cm_open (VenusFid fid, CredCacheEntry* ce, u_int tokens)
{
     FCacheEntry *entry;
     Result ret;
     u_long mask;
     int error;

     error = fcache_get (&entry, fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     assert (CheckLock (&entry->lock) == -1);

     error = fcache_get_data (entry, ce);
     if(error) {
	 ReleaseWriteLock (&entry->lock);
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     switch(tokens) {
     case XFS_DATA_R:
       /*     case XFS_OPEN_NR:
     case XFS_OPEN_SR: */
	  mask = AREAD;
	  break;
     case XFS_DATA_W:
	  mask = AWRITE;
	  break;
     case XFS_OPEN_NW:
	  mask = AREAD | AWRITE;
	  break;
     default:
	 ARLADEB (ADEBMISC, ("cm_open(): unknown token: %d, assuming AREAD\n",
		 tokens));
	 mask = AREAD;
/*	  assert(FALSE); */
     }

     if (checkright (entry, mask, ce)) {
	  if (mask == AREAD)
	       ++entry->nreaders;
	  else
	       ++entry->nwriters;
	  ret.res = entry->inode;
	  entry->flags.datausedp = TRUE;
	  entry->tokens |= tokens;

	  ret.res = entry->inode;
	  ret.tokens = entry->tokens;
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
     ReleaseWriteLock (&entry->lock);

     return ret;
}

/*
 * close. Set flags and if we opened the file for writing, write it
 * back to the server.
 */

Result
cm_close (VenusFid fid, int flag, CredCacheEntry* ce)
{
    FCacheEntry *entry;
    Result ret;
    int error;

    error = fcache_get (&entry, fid, ce);
    if (error) {
	ret.res = -1;
	ret.errno = error;
	return ret;
    }

    if (flag & XFS_WRITE && write_data (entry, ce)) {
	ReleaseWriteLock (&entry->lock);
	ARLADEB (ADEBMISC, ("Problems writing back file\n"));
	ret.res = -1 ;
	ret.errno = EPERM ;
	return ret;
    }
    if (flag & XFS_WRITE)
	--entry->nwriters;
    else if(flag & XFS_READ)
	--entry->nreaders;
    
    if (entry->nreaders == 0 && entry->nwriters == 0)
	entry->flags.datausedp = FALSE;
    
    ReleaseWriteLock (&entry->lock);
    ret.res = 0;
    return ret;
}

/*
 * getattr - read the attributes from this file.
 */

Result
cm_getattr (VenusFid fid,
	    AFSFetchStatus *attr,
	    VenusFid *realfid,
	    CredCacheEntry* ce)
{
     FCacheEntry *entry;
     Result ret;
     int error;

     error = fcache_get (&entry, fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (entry, ce);
     if (error) {
	 ReleaseWriteLock (&entry->lock);
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     if (checkright (entry,
		     entry->status.FileType == TYPE_FILE ? AREAD : ALIST,
		     ce)) {
	  *attr = entry->status;
	  ret.res = 0;
	  entry->flags.attrusedp = TRUE;
	  entry->tokens |= XFS_ATTR_R;
     } else {
	  ret.res   = -1;
	  ret.errno = EACCES;
     }
     if(entry->flags.mountp)
	 *realfid = entry->realfid;
     else
	 *realfid = fid;

     ReleaseWriteLock (&entry->lock);
     ret.tokens = entry->tokens;
     return ret;
}

/*
 * setattr - set the attributes of this file. These are immediately
 * sent to the FS.
 */

Result
cm_setattr (VenusFid fid, AFSStoreStatus *attr, CredCacheEntry* ce)
{
     FCacheEntry *entry;
     Result ret;
     int error;

     error = fcache_get (&entry, fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (entry, ce);
     if (error) {
	 ReleaseWriteLock (&entry->lock);
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }
     if (checkright (entry, AWRITE, ce)) {
	  ARLADEB (ADEBMISC, ("cm_setattr: Writing status\n"));
	  ret.res = write_attr (entry, attr, ce);
	  if (ret.res != 0)
	      ret.errno = ret.res;
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
     ReleaseWriteLock (&entry->lock);
     return ret;
}

/*
 * ftruncate - make the specified file have a specified size
 */

Result
cm_ftruncate (VenusFid fid, off_t size, CredCacheEntry* ce)
{
     FCacheEntry *entry;
     Result ret;
     int error;

     error = fcache_get (&entry, fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (entry, ce);
     if (error) {
	 ReleaseWriteLock (&entry->lock);
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     if (checkright (entry, AWRITE, ce)) {
	  ret.res = truncate_file (entry, size, ce);
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
     ReleaseWriteLock (&entry->lock);
     return ret;
}

/*
 * access - check if user is allowed to perform operation.
 */

Result
cm_access (VenusFid fid, int mode, CredCacheEntry* ce)
{
     FCacheEntry *entry;
     Result ret;
     int error;

     error = fcache_get (&entry, fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (entry, ce);
     if (error) {
	 ReleaseWriteLock (&entry->lock);
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     if (checkright (entry, AWRITE, ce)) {
	  ret.res = 0;		/**/
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
     ReleaseWriteLock (&entry->lock);
     return ret;
}

/*
 * Find this entry in the directory. If the entry happens to point to
 * a mount point, then we follow that and return the root directory of
 * the volume. Hopefully this is the only place where we need to think
 * about mount points.
 */

Result
cm_lookup (VenusFid fid, char *name, VenusFid *res, CredCacheEntry** ce)
{
     Result ret;
     int error;

     if (strcmp(name, "@sys") == 0)
	 name = arlasysname;

     error = adir_lookup (fid, name, res, *ce);
     if (error) {
	  ret.res = -1;
	  ret.errno = error;
	  return ret;
     }
     error = followmountpoint (res, &fid, ce);
     if (error) {
	  ret.res = -1;
	  ret.errno = error;
     } else {
	  ret.res = 0;
	  ret.tokens = 0;
     }

     /* Assume that this means a bad .. */
     if (   strcmp("..", name) == 0 
	 && fid.Cell == res->Cell 
	 && fid.fid.Volume == res->fid.Volume
	 && fid.fid.Vnode == res->fid.Vnode 
	 && fid.fid.Unique == res->fid.Unique) {
	  FCacheEntry *e;
	  int error;

	  error = fcache_get (&e, fid, *ce);
	  if (error) {
	      ret.res = -1;
	      ret.errno = error;
	      return ret;
	  }

	  error = fcache_get_attr (e, *ce);
	  if (error) {
	      ReleaseWriteLock (&e->lock);
	      ret.res = -1;
	      ret.errno = error;
	      return ret;
	  }

	  if (e->flags.mountp) {
	       *res = e->parent;
	       ret.res = 0;
	       ret.tokens = e->tokens;
	  } else {
	       ret.res = -1;
	       ret.errno = ENXIO;
	       ARLADEB (ADEBMISC, ("cm_lookup: bad ..\n"));
	  }
	  ReleaseWriteLock (&e->lock);
     }

     return ret;
}

/*
 * Create this file and more.
 */

Result
cm_create (VenusFid fid, char *name, AFSStoreStatus *store_attr,
	   VenusFid *res, AFSFetchStatus *fetch_attr,
	   CredCacheEntry* ce)
{
     FCacheEntry *dire;
     Result ret;
     int error;

     error = fcache_get (&dire, fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (dire, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 goto out;
     }

     if (checkright (dire, AINSERT, ce)) {
	  error = create_file (dire, name, store_attr,
			       res, fetch_attr, ce);
	  if (error) {
	      ret.res = -1;
	      ret.errno = error;
	  } else {
	      error = adir_creat (dire, name, res->fid);
	      if (error) {
		  ret.res = -1;
		  ret.errno = error;
	      } else {
		  ret.res = 0;
	      }
	  }
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
out:
     ReleaseWriteLock (&dire->lock);
     return ret;
}

/*
 * Create a new directory
 */

Result
cm_mkdir (VenusFid fid, char *name, AFSStoreStatus *store_attr,
	  VenusFid *res, AFSFetchStatus *fetch_attr,
	  CredCacheEntry* ce)
{
     FCacheEntry *dire;
     Result ret;
     int error;

     error = fcache_get (&dire, fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (dire, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 goto out;
     }

     if (checkright (dire, AINSERT, ce)) {
	  error = create_directory (dire, name, store_attr,
				    res, fetch_attr, ce);
	  if (error) {
	      ret.res = -1;
	      ret.errno = error;
	  } else {
	      error = adir_creat (dire, name, res->fid);
	      if (error) {
		  ret.res = -1;
		  ret.errno = error;
	      } else {
		  ret.res = 0;
	      }
	  }
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
out:
     ReleaseSharedLock (&dire->lock);
     return ret;
}

/*
 * Create a symlink
 */

Result
cm_symlink (VenusFid fid, char *name, AFSStoreStatus *store_attr,
	    VenusFid *res, AFSFetchStatus *fetch_attr,
	    char *contents,
	    CredCacheEntry* ce)
{
     FCacheEntry *dire;
     Result ret;
     int error;

     error = fcache_get (&dire, fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (dire, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 goto out;
     }

     if (checkright (dire, AINSERT, ce)) {
	  error = create_symlink (dire, name, store_attr,
				  res, fetch_attr,
				  contents, ce);
	  if (error) {
	      ret.res = -1;
	      ret.errno = error;
	  } else {
	      error = adir_creat (dire, name, res->fid);
	      if (error) {
		  ret.res = -1;
		  ret.errno = error;
	      } else {
		  ret.res = 0;
	      }
	  }
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
out:
     ReleaseWriteLock (&dire->lock);
     return ret;
}

/*
 * Create a hard link.
 */

Result
cm_link (VenusFid dir_fid, char *name,
	 VenusFid existing_fid,
	 AFSFetchStatus *existing_status,
	 CredCacheEntry* ce)
{
     FCacheEntry *dire, *file;
     Result ret;
     int error;

     error = fcache_get (&dire, dir_fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (dire, ce);
     if (error) {
	 ReleaseWriteLock (&dire->lock);
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get (&file, existing_fid, ce);
     if (error) {
	 ReleaseWriteLock (&dire->lock);
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (file, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 goto out;
     }

     if (checkright (dire, AINSERT, ce)) {
	  error = create_link (dire, name, 
			       file, ce);
	  if (error) {
	      ret.res = -1;
	      ret.errno = error;
	  } else {
	      error = adir_creat (dire, name, existing_fid.fid);
	      if (error) {
		  ret.res = -1;
		  ret.errno = error;
	      } else {
		  *existing_status = file->status;
		  ret.res = 0;
	      }
	  }
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
out:
     ReleaseWriteLock (&dire->lock);
     ReleaseWriteLock (&file->lock);
     return ret;
}
/*
 *
 */

Result
cm_remove(VenusFid fid, char *name, CredCacheEntry* ce)
{
     FCacheEntry *dire;
     Result ret;
     int error;

     error = fcache_get (&dire, fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (dire, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 goto out;
     }

     if (checkright (dire, ADELETE, ce)) {
	  error = remove_file (dire, name, ce);
	  if (error) {
	      ret.res = -1;
	      ret.errno = error;
	  } else {
	      error = adir_remove (dire, name);
	      if (error) {
		  ret.res = -1;
		  ret.errno = error;
	      } else {
		  ret.res = 0;
	      }
	  }
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
out:
     ReleaseWriteLock (&dire->lock);
     return ret;
}

/*
 *
 */

Result
cm_rmdir(VenusFid fid, char *name, CredCacheEntry* ce)
{
     FCacheEntry *dire;
     Result ret;
     int error;

     error = fcache_get (&dire, fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (dire, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 goto out;
     }

     if (checkright (dire, ADELETE, ce)) {
	  error = remove_directory (dire, name, ce);
	  if (error) {
	      ret.res = -1;
	      ret.errno = error;
	  } else {
	      error = adir_remove (dire, name);
	      if (error) {
		  ret.res = -1;
		  ret.errno = error;
	      } else {
		  ret.res = 0;
	      }
	  }
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
out:
     ReleaseWriteLock (&dire->lock);
     return ret;
}

/*
 *
 */

Result
cm_rename(VenusFid old_parent_fid, char *old_name,
	  VenusFid new_parent_fid, char *new_name,
	  CredCacheEntry* ce)
{
     FCacheEntry *old_dir;
     FCacheEntry *new_dir;
     Result ret;
     int error;

     /* old parent dir */

     error = fcache_get (&old_dir, old_parent_fid, ce);
     if (error) {
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     error = fcache_get_attr (old_dir, ce);
     if (error) {
	 ReleaseWriteLock (&old_dir->lock);
	 ret.res = -1;
	 ret.errno = error;
	 return ret;
     }

     /* new parent dir */

     if (old_parent_fid.fid.Vnode != new_parent_fid.fid.Vnode
	 || old_parent_fid.fid.Unique != new_parent_fid.fid.Unique) {

	 error = fcache_get (&new_dir, new_parent_fid, ce);
	 if (error) {
	     ReleaseWriteLock (&old_dir->lock);
	     ret.res = -1;
	     ret.errno = error;
	     return ret;
	 }

	 error = fcache_get_attr (new_dir, ce);
	 if (error) {
	     ret.res = -1;
	     ret.errno = error;
	     goto out;
	 }

     } else {
	 new_dir = old_dir;
     }

     if (checkright (old_dir, ADELETE, ce)
	 && checkright (new_dir, AINSERT, ce)) {

	 error = rename_file (old_dir, old_name,
			      new_dir, new_name,
			      ce);
	  if (error) {
	      ret.res = -1;
	      ret.errno = error;
	  } else {
	      VenusFid foo_fid;

	      ReleaseWriteLock (&old_dir->lock);
	      error = adir_lookup (old_dir->fid, old_name, &foo_fid, ce);
	      ObtainWriteLock (&old_dir->lock);

	      if (error) {
		  ret.res = -1;
		  ret.errno = errno;
		  goto out;
	      }
	      error = adir_remove (old_dir, old_name)
		  || adir_creat (new_dir, new_name, foo_fid.fid);

	      if (error) {
		  ret.res = -1;
		  ret.errno = error;
	      } else {
		  ret.res = 0;
	      }
	  }
     } else {
	  ret.res = -1;
	  ret.errno = EACCES;
     }
out:
     ReleaseWriteLock (&old_dir->lock);
     if (old_parent_fid.fid.Vnode != new_parent_fid.fid.Vnode
	 || old_parent_fid.fid.Unique != new_parent_fid.fid.Unique)
	 ReleaseWriteLock (&new_dir->lock);
     return ret;
}
