/*
 * P_R_P_Q_# (C) COPYRIGHT IBM CORPORATION 1987
 * LICENSED MATERIALS - PROPERTY OF IBM
 * REFER TO COPYRIGHT INSTRUCTIONS FORM NUMBER G120-2083
 */
/* getaddr -- get our internet address
David Nichols
July, 1986 */

#include "rx_locl.h"

RCSID("$Id: rx_gaddr.c,v 1.5 1998/02/22 19:44:51 joda Exp $");

#define NIFS		3

#ifdef KERNEL
/* only used for generating random noise */

long rxi_tempAddr = 0;		       /* default attempt */

/* set the advisory noise */
long 
rxi_setaddr(long x)
{
    rxi_tempAddr = x;
}

/* get approx to net addr */
long 
rxi_getaddr(void)
{
    return rxi_tempAddr;
}

#else				       /* KERNEL */

/* to satisfy those who call setaddr */
int
rxi_setaddr(long x)
{
    return 0;
}

/*
 * Return our internet address as a long in network byte order.  Returns zero
 * if it can't find one.
 */
long 
rxi_getaddr(void)
{
    int s;
    int i, len;
    struct ifconf ifc;
    struct ifreq ifs[NIFS];
    struct sockaddr_in *a;

    s = socket(AF_INET, SOCK_DGRAM, 0);
    if (s < 0)
	return 0;
    ifc.ifc_len = sizeof(ifs);
    ifc.ifc_buf = (caddr_t) ifs;
    i = ioctl(s, SIOCGIFCONF, &ifc);
    close(s);
    if (i < 0)
	return 0;
    len = ifc.ifc_len / sizeof(struct ifreq);
    if (len > NIFS)
	len = NIFS;
    for (i = 0; i < len; ++i) {
	a = (struct sockaddr_in *) & ifs[i].ifr_addr;
	if (a->sin_addr.s_addr != 0 && strcmp(ifs[i].ifr_name, "lo0") != 0)
	    return a->sin_addr.s_addr;
    }
    return 0;
}

#endif				       /* KERNEL */
